#include "minigpsd.h"

static char xpmimage[4096];
static char gpscirc[40 * 40 + 1] =
  "  A            ++++++++++               "
  "  AA        ++++++++++++++++            "
  "A A A     ++++++----====++++++          "
  " AAA     ++++-------=======++++         "
  "A A A  ++++---------=========++++       "
  "  AA  +++-----------===========+++      "
  "  A  +++------------============+++     "
  "    +++-------------=============+++    "
  "    ++----------====----==========++    "
  "   ++---------======------=========++   "
  "  +++--------=======-------========+++  "
  "  ++--------========--------========++  "
  " +++-------=========---------=======+++ "
  " ++-------==========----------=======++ "
  " ++------=========--==---------======++ "
  "+++------=======----====-------======+++"
  "++------=======-----=====-------======++"
  "++------=======-----=====-------======++"
  "++------======------======------======++"
  "++------======------======------======++"
  "++======------======------======------++"
  "++======------======------======------++"
  "++======-------=====-----=======------++"
  "++======-------=====-----=======------++"
  "+++======-------====----=======------+++"
  " ++======---------==--=========------++ "
  " ++=======----------==========-------++ "
  " +++=======---------=========-------+++ "
  "  ++========--------========--------++  "
  "  +++========-------=======--------+++  "
  "   ++=========------======---------++   "
  "    ++==========----====----------++    "
  "    +++=============-------------+++    "
  "  B  +++============------------+++     "
  "  BB  +++===========-----------+++      "
  "B B B  ++++=========---------++++       "
  " BBB     ++++=======-------++++         "
  "B B B     ++++++====----++++++          "
  "  BB        ++++++++++++++++            "
  "  B            ++++++++++               ";

static char gpscircedit[40 * 40 + 1];

int sinetab[] =
  { 0, 571, 1143, 1714, 2285, 2855, 3425, 3993, 4560, 5126, 5690, 6252, 6812, 7371, 7927, 8480, 9032, 9580, 10125, 10668, 11207,
    11743, 12275, 12803, 13327, 13848, 14364, 14876, 15383, 15886, 16383, 16876, 17364, 17846, 18323, 18794, 19260, 19720,
    20173, 20621, 21062, 21497, 21926, 22347, 22762, 23170, 23571, 23964, 24351, 24730, 25101, 25465, 25821, 26169, 26509, 26841,
    27165, 27481, 27788, 28087, 28377, 28659, 28932, 29196, 29451, 29697, 29935, 30163, 30381, 30591, 30791, 30982, 31164,
    31336, 31498, 31651, 31794, 31928, 32051, 32165, 32270, 32364, 32449, 32523, 32588, 32643, 32688, 32723, 32748, 32763, 32768
};

int isin(angle)
{
    while (angle >= 360)
        angle -= 360;
    while (angle < 0)
        angle += 360;
    if (angle <= 90)
        return sinetab[angle];

    else if (angle <= 180)
        return sinetab[180 - angle];

    else if (angle <= 270)
        return -sinetab[angle - 180];

    else
        return -sinetab[360 - angle];
};

int icos(int angle)
{
    return isin(angle + 90);
};

static int lastsec, lastmn;

static int tlx, tly;
static void aetoxy(azim, elev)
{
    tlx = 20 + (90 - elev) * (isin(azim) * 18 / 32768) / 90;
    tly = 20 - (90 - elev) * (isin(90 - azim) * 18 / 32768) / 90;
};

extern char gpsgateurl[];
extern int gpsgatefd;
void dogpsgate(void);

extern int iconupdaterate;
static int iconupdatelast = 0;

extern int obdfd;
extern int gpsfd;

static FILE *xpmfp = NULL;
void xpmgpsstatus()
{
    int ncolors = 16, h = 40, w = 40, i;
    char *gc = gpscircedit;
    char *c = xpmimage;
    int n;

    if (sc == lastsec)
        return;
    lastsec = sc;

    if( iconupdatelast-- )
    	return;

    iconupdatelast = iconupdaterate;
    
    if (lastmn != mn) {
        lastmn = mn;
        if (gpsgateurl[0] && gpsgatefd < 0) {
            dogpsgate();
            return;             //don't spend more time
        }
    }

    strcpy(gc, gpscirc);

    if (gspd > 3000) {          // moving at least 3mph
        aetoxy(gtrk / 1000, -5);
        //fprintf(stderr, "Sp%d Gt%d %d,%d\n", gspd, gtrk, tlx, tly);
        gc[tly * w + tlx] = '9';
        gc[(tly - 1) * w + tlx] = '9';
        gc[(tly - 1) * w + tlx - 1] = '9';
        gc[tly * w + tlx - 1] = '9';
        gc[(tly + 1) * w + tlx] = '9';
        gc[(tly - 2) * w + tlx] = '9';
        gc[(tly - 2) * w + tlx - 1] = '9';
        gc[(tly + 1) * w + tlx - 1] = '9';
        gc[tly * w + tlx + 1] = '9';
        gc[(tly - 1) * w + tlx + 1] = '9';
        gc[(tly - 1) * w + tlx - 2] = '9';
        gc[tly * w + tlx - 2] = '9';
    }

    for (n = 0; n < nsats; n++) {
        int m, k, sc;
        m = view[n];
        sc = sn[m];
        sc /= 3;
        sc -= 5;
        if (sc < 0)
            sc = 0;
        if (sc > 8)
            sc = 8;
        sc += '0';
        for (k = 0; k < 12; k++)
            if (sats[k] == m)
                break;
        aetoxy(az[m], el[m]);
        gc[tly * w + tlx] = sc;
        gc[(tly - 1) * w + tlx] = sc;
        gc[(tly - 1) * w + tlx - 1] = sc;
        gc[tly * w + tlx - 1] = sc;
        if (k < 12) {
            gc[(tly + 1) * w + tlx] = sc;
            gc[(tly - 2) * w + tlx] = sc;
            gc[(tly - 2) * w + tlx - 1] = sc;
            gc[(tly + 1) * w + tlx - 1] = sc;
            gc[tly * w + tlx + 1] = sc;
            gc[(tly - 1) * w + tlx + 1] = sc;
            gc[(tly - 1) * w + tlx - 2] = sc;
            gc[tly * w + tlx - 2] = sc;
        }
        //        fprintf(stderr, "%02d %d %d  %d,%d %c\n", m, el[m], az[m], tlx, tly, sc);
    }
    sprintf(xpmimage, "/* XPM */\nstatic char *gpsxpm[] = {\n");
    c += strlen(xpmimage);
    sprintf(c, "\"40 40 %d 1\",\n", ncolors);
    c += strlen(c);
    sprintf(c, "\"  c None\",\n\"- c #222222\",\n\"= c #444444\",\n");
    c += strlen(c);

    int r = 255, g = 0, b = 0;
    if (lock > 1)
        r = 0, b = 255;
    else if (lock)
	g = 255;
	if (fix == 3)
	    r = 0;

    sprintf(c, "\"+ c #%02X%02X%02x\",\n", r, g, b);
    c += strlen(c);

    sprintf(c, "\"0 c #ff0000\",\n"
	    "\"1 c #ff8000\",\n"
	    "\"2 c #ffc000\",\n"
	    "\"3 c #ffff00\",\n"
	    "\"4 c #80ff00\",\n"
	    "\"5 c #00ff00\",\n"
	    "\"6 c #80ff80\",\n"
	    "\"7 c #ffffff\",\n"
	    "\"8 c #ff00ff\",\n"
	    "\"9 c #008080\",\n");
    c += strlen(c);

    r = gpsfd > -1 ? 255 : 0;
    g = obdfd > -1 ? 255 : 0;
    sprintf( c, "\"A c #%02X%02X%02X\",\n"
	     "\"B c #%02X%02X%02X\",\n",
	     0,r,0,0,g,0
	     );

    c += strlen(c);

    for (i = 0; i < h; i++) {
        sprintf(c, "\"%40.40s\",\n", gc);
        c += strlen(c);
        gc += w;
    }
    sprintf(c, "};\n");
    rewind(xpmfp);
    fwrite(xpmimage, strlen(xpmimage), 1, xpmfp);
}

void xpmgpsinit(int opcl)
{
    int ncolors = 16, h = 40, w = 40, i;
    char *gc = gpscircedit;
    char *c = xpmimage;
    strcpy(gc, gpscirc);

    if (opcl) {
        unlink("/tmp/gps.xpm");
        xpmfp = fopen("/tmp/gps.xpm", "w+");
    }

    sprintf(xpmimage, "/* XPM */\nstatic char *gpsxpm[] = {\n");
    c += strlen(xpmimage);
    sprintf(c, "\"40 40 %d 1\",\n", ncolors);
    c += strlen(c);
    sprintf(c, "\"  c None\",\n");
    c += strlen(c);

    if (opcl)
        sprintf(c, "\"- c #888888\",\n" "\"= c #cccccc\",\n" "\"+ c #ffffff\",\n");
    else
        sprintf(c, "\"- c #333333\",\n" "\"= c #666666\",\n" "\"+ c #000000\",\n");
    c += strlen(c);

    sprintf(c, "\"0 c #0000ff\",\n\"1 c #cc0000\",\n\"2 c #ff8000\",\n\"3 c #ffa000\",\n\"4 c #ffc000\",\n"
"\"A c #000000\",\n\"B c #000000\",\n"
      "\"5 c #ffff00\",\n\"6 c #ffff00\",\n\"7 c #ffff80\",\n\"8 c #ffffff\",\n\"9 c #00ffff\",\n");
    c += strlen(c);

    for (i = 0; i < h; i++) {
        sprintf(c, "\"%40.40s\",\n", gc);
        c += strlen(c);
        gc += w;
    }
    sprintf(c, "};\n");

    rewind(xpmfp);
    fwrite(xpmimage, strlen(xpmimage), 1, xpmfp);

    if (!opcl)
        fclose(xpmfp);
}
